@@ -1,6 +1,6 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
-  Copyright(C) 2009-2015 Brazil
+  Copyright(C) 2009-2016 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -23,6 +23,7 @@
 #include "grn_ctx_impl.h"
 #include "grn_ii.h"
 #include "grn_pat.h"
+#include "grn_index_column.h"
 #include "grn_proc.h"
 #include "grn_plugin.h"
 #include "grn_snip.h"
@@ -35,6 +36,14 @@
 #include <stdarg.h>
 #include <time.h>
 
+#ifdef GRN_WITH_ONIGMO
+# define GRN_SUPPORT_REGEXP
+#endif /* GRN_WITH_ONIGMO */
+
+#ifdef GRN_SUPPORT_REGEXP
+# include <oniguruma.h>
+#endif /* GRN_SUPPORT_REGEXP */
+
 #ifdef WIN32
 # include <share.h>
 #else /* WIN32 */
@@ -96,10 +105,27 @@ grn_init_from_env(void)
   grn_io_init_from_env();
   grn_ii_init_from_env();
   grn_db_init_from_env();
+  grn_index_column_init_from_env();
   grn_proc_init_from_env();
   grn_plugin_init_from_env();
 }
 
+static void
+grn_init_external_libraries(void)
+{
+#ifdef GRN_SUPPORT_REGEXP
+  onig_init();
+#endif /*  GRN_SUPPORT_REGEXP */
+}
+
+static void
+grn_fin_external_libraries(void)
+{
+#ifdef GRN_SUPPORT_REGEXP
+  onig_end();
+#endif /*  GRN_SUPPORT_REGEXP */
+}
+
 void
 grn_sleep(uint32_t seconds)
 {
@@ -189,7 +215,7 @@ grn_timeval2tm(grn_ctx *ctx, grn_timeval *tv, struct tm *tm_buffer)
 # endif /* HAVE_LOCALTIME_R */
 #endif /* HAVE__LOCALTIME64_S */
   if (!ltm) {
-    SERR(function_name);
+    SERR("%s", function_name);
   }
   return ltm;
 }
@@ -222,8 +248,9 @@ grn_str2timeval(const char *str, uint32_t str_len, grn_timeval *tv)
   memset(&tm, 0, sizeof(struct tm));
 
   tm.tm_year = (int)grn_atoui(str, rend, &r1) - 1900;
-  if ((r1 + 1) >= rend || (*r1 != '/' && *r1 != '-') ||
-      tm.tm_year < 0) { return GRN_INVALID_ARGUMENT; }
+  if ((r1 + 1) >= rend || (*r1 != '/' && *r1 != '-')) {
+    return GRN_INVALID_ARGUMENT;
+  }
   r1++;
   tm.tm_mon = (int)grn_atoui(r1, rend, &r1) - 1;
   if ((r1 + 1) >= rend || (*r1 != '/' && *r1 != '-') ||
@@ -268,6 +295,12 @@ grn_str2timeval(const char *str, uint32_t str_len, grn_timeval *tv)
   return GRN_SUCCESS;
 }
 
+const char *
+grn_get_global_error_message(void)
+{
+  return grn_gctx.errbuf;
+}
+
 #ifdef USE_MEMORY_DEBUG
 inline static void
 grn_alloc_info_set_backtrace(char *buffer, size_t size)
@@ -509,17 +542,17 @@ static int
 grn_msgpack_buffer_write(void *data, const char *buf, msgpack_size_t len)
 {
   grn_ctx *ctx = (grn_ctx *)data;
-  return grn_bulk_write(ctx, ctx->impl->outbuf, buf, len);
+  return grn_bulk_write(ctx, ctx->impl->output.buf, buf, len);
 }
 #endif
 
-static void
+static grn_rc
 grn_ctx_impl_init(grn_ctx *ctx)
 {
   grn_io_mapinfo mi;
   if (!(ctx->impl = grn_io_anon_map(ctx, &mi, IMPL_SIZE))) {
     ctx->impl = NULL;
-    return;
+    return ctx->rc;
   }
 #ifdef USE_DYNAMIC_MALLOC_CHANGE
   grn_ctx_impl_init_malloc(ctx);
@@ -536,16 +569,27 @@ grn_ctx_impl_init(grn_ctx *ctx)
     CRITICAL_SECTION_FIN(ctx->impl->lock);
     grn_io_anon_unmap(ctx, &mi, IMPL_SIZE);
     ctx->impl = NULL;
-    return;
+    return ctx->rc;
+  }
+  if (!(ctx->impl->temporary_columns = grn_pat_create(ctx, NULL,
+                                                      GRN_TABLE_MAX_KEY_SIZE,
+                                                      sizeof(grn_obj *),
+                                                      0))) {
+    grn_array_close(ctx, ctx->impl->values);
+    CRITICAL_SECTION_FIN(ctx->impl->lock);
+    grn_io_anon_unmap(ctx, &mi, IMPL_SIZE);
+    ctx->impl = NULL;
+    return ctx->rc;
   }
   if (!(ctx->impl->ios = grn_hash_create(ctx, NULL, GRN_TABLE_MAX_KEY_SIZE,
                                          sizeof(grn_io *),
                                          GRN_OBJ_KEY_VAR_SIZE|GRN_HASH_TINY))) {
     grn_array_close(ctx, ctx->impl->values);
+    grn_pat_close(ctx, ctx->impl->temporary_columns);
     CRITICAL_SECTION_FIN(ctx->impl->lock);
     grn_io_anon_unmap(ctx, &mi, IMPL_SIZE);
     ctx->impl = NULL;
-    return;
+    return ctx->rc;
   }
   ctx->impl->db = NULL;
 
@@ -555,8 +599,8 @@ grn_ctx_impl_init(grn_ctx *ctx)
   ctx->impl->qe_next = NULL;
   ctx->impl->parser = NULL;
 
-  GRN_TEXT_INIT(&ctx->impl->names, GRN_OBJ_VECTOR);
-  GRN_UINT32_INIT(&ctx->impl->levels, GRN_OBJ_VECTOR);
+  GRN_TEXT_INIT(&ctx->impl->output.names, GRN_OBJ_VECTOR);
+  GRN_UINT32_INIT(&ctx->impl->output.levels, GRN_OBJ_VECTOR);
 
   if (ctx == &grn_gctx) {
     ctx->impl->command_version = GRN_COMMAND_VERSION_STABLE;
@@ -575,9 +619,13 @@ grn_ctx_impl_init(grn_ctx *ctx)
   ctx->impl->finalizer = NULL;
 
   ctx->impl->com = NULL;
-  ctx->impl->outbuf = grn_obj_open(ctx, GRN_BULK, 0, GRN_DB_TEXT);
-  ctx->impl->output = NULL;
-  ctx->impl->data.ptr = NULL;
+  ctx->impl->output.buf = grn_obj_open(ctx, GRN_BULK, 0, GRN_DB_TEXT);
+  ctx->impl->output.func = NULL;
+  ctx->impl->output.data.ptr = NULL;
+#ifdef GRN_WITH_MESSAGE_PACK
+  msgpack_packer_init(&ctx->impl->output.msgpacker,
+                      ctx, grn_msgpack_buffer_write);
+#endif
   ctx->impl->tv.tv_sec = 0;
   ctx->impl->tv.tv_nsec = 0;
   ctx->impl->edge = NULL;
@@ -589,11 +637,9 @@ grn_ctx_impl_init(grn_ctx *ctx)
   ctx->impl->previous_errbuf[0] = '\0';
   ctx->impl->n_same_error_messages = 0;
 
-#ifdef GRN_WITH_MESSAGE_PACK
-  msgpack_packer_init(&ctx->impl->msgpacker, ctx, grn_msgpack_buffer_write);
-#endif
-
   grn_ctx_impl_mrb_init(ctx);
+
+  return ctx->rc;
 }
 
 void
@@ -732,6 +778,15 @@ grn_ctx_fin(grn_ctx *ctx)
 #endif
       grn_array_close(ctx, ctx->impl->values);
     }
+    if (ctx->impl->temporary_columns) {
+#ifndef USE_MEMORY_DEBUG
+      grn_obj *value;
+      GRN_PAT_EACH(ctx, ctx->impl->temporary_columns, id, NULL, NULL, &value, {
+        grn_obj_close(ctx, *((grn_obj **)value));
+      });
+#endif
+      grn_pat_close(ctx, ctx->impl->temporary_columns);
+    }
     if (ctx->impl->ios) {
       grn_hash_close(ctx, ctx->impl->ios);
     }
@@ -746,10 +801,10 @@ grn_ctx_fin(grn_ctx *ctx)
       grn_ctx_send(ctx, "ACK", 3, GRN_CTX_HEAD);
       rc = grn_com_close(ctx, ctx->impl->com);
     }
-    GRN_OBJ_FIN(ctx, &ctx->impl->names);
-    GRN_OBJ_FIN(ctx, &ctx->impl->levels);
     GRN_OBJ_FIN(ctx, &ctx->impl->query_log_buf);
-    rc = grn_obj_close(ctx, ctx->impl->outbuf);
+    GRN_OBJ_FIN(ctx, &ctx->impl->output.names);
+    GRN_OBJ_FIN(ctx, &ctx->impl->output.levels);
+    rc = grn_obj_close(ctx, ctx->impl->output.buf);
     {
       grn_hash **vp;
       grn_obj *value;
@@ -842,15 +897,22 @@ grn_init(void)
   grn_rc rc;
   grn_ctx *ctx = &grn_gctx;
   grn_init_from_env();
+  grn_init_external_libraries();
   grn_logger_init();
   grn_query_logger_init();
   CRITICAL_SECTION_INIT(grn_glock);
   grn_gtick = 0;
   ctx->next = ctx;
   ctx->prev = ctx;
-  grn_ctx_init_internal(ctx, 0);
+  rc = grn_ctx_init_internal(ctx, 0);
+  if (rc) {
+    goto fail_ctx_init_internal;
+  }
   ctx->encoding = grn_encoding_parse(GRN_DEFAULT_ENCODING);
-  grn_timeval_now(ctx, &grn_starttime);
+  rc = grn_timeval_now(ctx, &grn_starttime);
+  if (rc) {
+    goto fail_start_time;
+  }
 #ifdef WIN32
   {
     SYSTEM_INFO si;
@@ -860,7 +922,8 @@ grn_init(void)
 #else /* WIN32 */
   if ((grn_pagesize = sysconf(_SC_PAGESIZE)) == -1) {
     SERR("_SC_PAGESIZE");
-    return ctx->rc;
+    rc = ctx->rc;
+    goto fail_page_size;
   }
 #endif /* WIN32 */
   if (grn_pagesize & (grn_pagesize - 1)) {
@@ -916,38 +979,55 @@ grn_init(void)
 #endif /* USE_FAIL_MALLOC */
   if ((rc = grn_com_init())) {
     GRN_LOG(ctx, GRN_LOG_ALERT, "grn_com_init failed (%d)", rc);
-    return rc;
+    goto fail_com;
+  }
+  if ((rc = grn_ctx_impl_init(ctx))) {
+    GRN_LOG(ctx, GRN_LOG_ALERT, "grn_ctx_impl_init failed (%d)", rc);
+    goto fail_ctx_impl;
   }
-  grn_ctx_impl_init(ctx);
   if ((rc = grn_plugins_init())) {
     GRN_LOG(ctx, GRN_LOG_ALERT, "grn_plugins_init failed (%d)", rc);
-    return rc;
+    goto fail_plugins;
   }
   if ((rc = grn_normalizer_init())) {
     GRN_LOG(ctx, GRN_LOG_ALERT, "grn_normalizer_init failed (%d)", rc);
-    return rc;
+    goto fail_normalizer;
   }
   if ((rc = grn_tokenizers_init())) {
     GRN_LOG(ctx, GRN_LOG_ALERT, "grn_tokenizers_init failed (%d)", rc);
-    return rc;
+    goto fail_tokenizer;
   }
-  /*
-  if ((rc = grn_index_init())) {
-    GRN_LOG(ctx, GRN_LOG_ALERT, "index initialize failed (%d)", rc);
-    return rc;
-  }
-  */
   grn_cache_init();
   if (!grn_request_canceler_init()) {
     rc = ctx->rc;
-    grn_cache_fin();
     GRN_LOG(ctx, GRN_LOG_ALERT,
             "failed to initialize request canceler (%d)", rc);
-    return rc;
+    goto fail_request_canceler;
   }
-  GRN_LOG(ctx, GRN_LOG_NOTICE, "grn_init");
+  GRN_LOG(ctx, GRN_LOG_NOTICE, "grn_init: <%s>", grn_get_version());
   check_overcommit_memory(ctx);
   return rc;
+
+fail_request_canceler:
+  grn_cache_fin();
+fail_tokenizer:
+  grn_normalizer_fin();
+fail_normalizer:
+  grn_plugins_fin();
+fail_plugins:
+  grn_ctx_fin(ctx);
+fail_ctx_impl:
+  grn_com_fin();
+fail_com:
+fail_page_size:
+fail_start_time:
+fail_ctx_init_internal:
+  GRN_LOG(ctx, GRN_LOG_NOTICE, "grn_init: <%s>: failed", grn_get_version());
+  grn_query_logger_fin(ctx);
+  grn_logger_fin(ctx);
+  CRITICAL_SECTION_FIN(grn_glock);
+  grn_fin_external_libraries();
+  return rc;
 }
 
 grn_encoding
@@ -1040,6 +1120,7 @@ grn_fin(void)
   GRN_LOG(ctx, GRN_LOG_NOTICE, "grn_fin (%d)", alloc_count);
   grn_logger_fin(ctx);
   CRITICAL_SECTION_FIN(grn_glock);
+  grn_fin_external_libraries();
   return GRN_SUCCESS;
 }
 
@@ -1100,7 +1181,7 @@ grn_content_type
 grn_ctx_get_output_type(grn_ctx *ctx)
 {
   if (ctx->impl) {
-    return ctx->impl->output_type;
+    return ctx->impl->output.type;
   } else {
     return GRN_CONTENT_NONE;
   }
@@ -1112,25 +1193,25 @@ grn_ctx_set_output_type(grn_ctx *ctx, grn_content_type type)
   grn_rc rc = GRN_SUCCESS;
 
   if (ctx->impl) {
-    ctx->impl->output_type = type;
-    switch (ctx->impl->output_type) {
+    ctx->impl->output.type = type;
+    switch (ctx->impl->output.type) {
     case GRN_CONTENT_NONE :
-      ctx->impl->mime_type = "application/octet-stream";
+      ctx->impl->output.mime_type = "application/octet-stream";
       break;
     case GRN_CONTENT_TSV :
-      ctx->impl->mime_type = "text/tab-separated-values";
+      ctx->impl->output.mime_type = "text/tab-separated-values";
       break;
     case GRN_CONTENT_JSON :
-      ctx->impl->mime_type = "application/json";
+      ctx->impl->output.mime_type = "application/json";
       break;
     case GRN_CONTENT_XML :
-      ctx->impl->mime_type = "text/xml";
+      ctx->impl->output.mime_type = "text/xml";
       break;
     case GRN_CONTENT_MSGPACK :
-      ctx->impl->mime_type = "application/x-msgpack";
+      ctx->impl->output.mime_type = "application/x-msgpack";
       break;
     case GRN_CONTENT_GROONGA_COMMAND_LIST :
-      ctx->impl->mime_type = "text/x-groonga-command-list";
+      ctx->impl->output.mime_type = "text/x-groonga-command-list";
       break;
     }
   } else {
@@ -1144,7 +1225,7 @@ const char *
 grn_ctx_get_mime_type(grn_ctx *ctx)
 {
   if (ctx->impl) {
-    return ctx->impl->mime_type;
+    return ctx->impl->output.mime_type;
   } else {
     return NULL;
   }
@@ -1195,70 +1276,70 @@ grn_get_ctype(grn_obj *var)
 static void
 get_content_mime_type(grn_ctx *ctx, const char *p, const char *pe)
 {
-  ctx->impl->output_type = GRN_CONTENT_NONE;
-  ctx->impl->mime_type = "application/octet-stream";
+  ctx->impl->output.type = GRN_CONTENT_NONE;
+  ctx->impl->output.mime_type = "application/octet-stream";
 
   if (p + 2 <= pe) {
     switch (*p) {
     case 'c' :
       if (p + 3 == pe && !memcmp(p, "css", 3)) {
-        ctx->impl->output_type = GRN_CONTENT_NONE;
-        ctx->impl->mime_type = "text/css";
+        ctx->impl->output.type = GRN_CONTENT_NONE;
+        ctx->impl->output.mime_type = "text/css";
       }
       break;
     case 'g' :
       if (p + 3 == pe && !memcmp(p, "gif", 3)) {
-        ctx->impl->output_type = GRN_CONTENT_NONE;
-        ctx->impl->mime_type = "image/gif";
+        ctx->impl->output.type = GRN_CONTENT_NONE;
+        ctx->impl->output.mime_type = "image/gif";
       }
       break;
     case 'h' :
       if (p + 4 == pe && !memcmp(p, "html", 4)) {
-        ctx->impl->output_type = GRN_CONTENT_NONE;
-        ctx->impl->mime_type = "text/html";
+        ctx->impl->output.type = GRN_CONTENT_NONE;
+        ctx->impl->output.mime_type = "text/html";
       }
       break;
     case 'j' :
       if (!memcmp(p, "js", 2)) {
         if (p + 2 == pe) {
-          ctx->impl->output_type = GRN_CONTENT_NONE;
-          ctx->impl->mime_type = "text/javascript";
+          ctx->impl->output.type = GRN_CONTENT_NONE;
+          ctx->impl->output.mime_type = "text/javascript";
         } else if (p + 4 == pe && !memcmp(p + 2, "on", 2)) {
-          ctx->impl->output_type = GRN_CONTENT_JSON;
-          ctx->impl->mime_type = "application/json";
+          ctx->impl->output.type = GRN_CONTENT_JSON;
+          ctx->impl->output.mime_type = "application/json";
         }
       } else if (p + 3 == pe && !memcmp(p, "jpg", 3)) {
-        ctx->impl->output_type = GRN_CONTENT_NONE;
-        ctx->impl->mime_type = "image/jpeg";
+        ctx->impl->output.type = GRN_CONTENT_NONE;
+        ctx->impl->output.mime_type = "image/jpeg";
       }
       break;
 #ifdef GRN_WITH_MESSAGE_PACK
     case 'm' :
       if (p + 7 == pe && !memcmp(p, "msgpack", 7)) {
-        ctx->impl->output_type = GRN_CONTENT_MSGPACK;
-        ctx->impl->mime_type = "application/x-msgpack";
+        ctx->impl->output.type = GRN_CONTENT_MSGPACK;
+        ctx->impl->output.mime_type = "application/x-msgpack";
       }
       break;
 #endif
     case 'p' :
       if (p + 3 == pe && !memcmp(p, "png", 3)) {
-        ctx->impl->output_type = GRN_CONTENT_NONE;
-        ctx->impl->mime_type = "image/png";
+        ctx->impl->output.type = GRN_CONTENT_NONE;
+        ctx->impl->output.mime_type = "image/png";
       }
       break;
     case 't' :
       if (p + 3 == pe && !memcmp(p, "txt", 3)) {
-        ctx->impl->output_type = GRN_CONTENT_NONE;
-        ctx->impl->mime_type = "text/plain";
+        ctx->impl->output.type = GRN_CONTENT_NONE;
+        ctx->impl->output.mime_type = "text/plain";
       } else if (p + 3 == pe && !memcmp(p, "tsv", 3)) {
-        ctx->impl->output_type = GRN_CONTENT_TSV;
-        ctx->impl->mime_type = "text/tab-separated-values";
+        ctx->impl->output.type = GRN_CONTENT_TSV;
+        ctx->impl->output.mime_type = "text/tab-separated-values";
       }
       break;
     case 'x':
       if (p + 3 == pe && !memcmp(p, "xml", 3)) {
-        ctx->impl->output_type = GRN_CONTENT_XML;
-        ctx->impl->mime_type = "text/xml";
+        ctx->impl->output.type = GRN_CONTENT_XML;
+        ctx->impl->output.mime_type = "text/xml";
       }
       break;
     }
@@ -1308,10 +1389,12 @@ get_command_version(grn_ctx *ctx, const char *p, const char *pe)
 #define OUTPUT_TYPE         "output_type"
 #define COMMAND_VERSION     "command_version"
 #define REQUEST_ID          "request_id"
+#define OUTPUT_PRETTY       "output_pretty"
 #define EXPR_MISSING        "expr_missing"
 #define OUTPUT_TYPE_LEN     (sizeof(OUTPUT_TYPE) - 1)
 #define COMMAND_VERSION_LEN (sizeof(COMMAND_VERSION) - 1)
 #define REQUEST_ID_LEN      (sizeof(REQUEST_ID) - 1)
+#define OUTPUT_PRETTY_LEN   (sizeof(OUTPUT_PRETTY) - 1)
 
 #define HTTP_QUERY_PAIR_DELIMITER   "="
 #define HTTP_QUERY_PAIRS_DELIMITERS "&;"
@@ -1326,9 +1409,13 @@ command_proc_p(grn_obj *expr)
 grn_obj *
 grn_ctx_qe_exec_uri(grn_ctx *ctx, const char *path, uint32_t path_len)
 {
+  grn_command_version command_version;
   grn_obj buf, *expr, *val;
   grn_obj request_id;
   const char *p = path, *e = path + path_len, *v, *key_end, *filename_end;
+
+  command_version = grn_ctx_get_command_version(ctx);
+
   GRN_TEXT_INIT(&buf, 0);
   GRN_TEXT_INIT(&request_id, 0);
   p = grn_text_urldec(ctx, &buf, p, e, '?');
@@ -1363,6 +1450,16 @@ grn_ctx_qe_exec_uri(grn_ctx *ctx, const char *path, uint32_t path_len)
           p = grn_text_cgidec(ctx, &request_id, p, e,
                               HTTP_QUERY_PAIRS_DELIMITERS);
           if (ctx->rc) { goto exit; }
+        } else if (l == OUTPUT_PRETTY_LEN &&
+                   !memcmp(v, OUTPUT_PRETTY, OUTPUT_PRETTY_LEN)) {
+          GRN_BULK_REWIND(&buf);
+          p = grn_text_cgidec(ctx, &buf, p, e, HTTP_QUERY_PAIRS_DELIMITERS);
+          if (GRN_TEXT_LEN(&buf) == strlen("yes") &&
+              !memcmp(GRN_TEXT_VALUE(&buf), "yes", GRN_TEXT_LEN(&buf))) {
+            ctx->impl->output.is_pretty = GRN_TRUE;
+          } else {
+            ctx->impl->output.is_pretty = GRN_FALSE;
+          }
         } else {
           if (!(val = grn_expr_get_or_add_var(ctx, expr, v, l))) {
             val = &buf;
@@ -1399,17 +1496,22 @@ grn_ctx_qe_exec_uri(grn_ctx *ctx, const char *path, uint32_t path_len)
 exit :
   GRN_OBJ_FIN(ctx, &request_id);
   GRN_OBJ_FIN(ctx, &buf);
+
+  grn_ctx_set_command_version(ctx, command_version);
+
   return expr;
 }
 
 grn_obj *
 grn_ctx_qe_exec(grn_ctx *ctx, const char *str, uint32_t str_len)
 {
+  grn_command_version command_version;
   char tok_type;
   int offset = 0;
   grn_obj buf, *expr = NULL, *val = NULL;
   grn_obj request_id;
   const char *p = str, *e = str + str_len, *v;
+  command_version = grn_ctx_get_command_version(ctx);
   GRN_TEXT_INIT(&buf, 0);
   GRN_TEXT_INIT(&request_id, 0);
   p = grn_text_unesc_tok(ctx, &buf, p, e, &tok_type);
@@ -1442,6 +1544,16 @@ grn_ctx_qe_exec(grn_ctx *ctx, const char *str, uint32_t str_len)
           GRN_BULK_REWIND(&request_id);
           p = grn_text_unesc_tok(ctx, &request_id, p, e, &tok_type);
           if (ctx->rc) { goto exit; }
+        } else if (l == OUTPUT_PRETTY_LEN &&
+                   !memcmp(v, OUTPUT_PRETTY, OUTPUT_PRETTY_LEN)) {
+          GRN_BULK_REWIND(&buf);
+          p = grn_text_unesc_tok(ctx, &buf, p, e, &tok_type);
+          if (GRN_TEXT_LEN(&buf) == strlen("yes") &&
+              !memcmp(GRN_TEXT_VALUE(&buf), "yes", GRN_TEXT_LEN(&buf))) {
+            ctx->impl->output.is_pretty = GRN_TRUE;
+          } else {
+            ctx->impl->output.is_pretty = GRN_FALSE;
+          }
         } else if (expr && (val = grn_expr_get_or_add_var(ctx, expr, v, l))) {
           grn_obj_reinit(ctx, val, GRN_DB_TEXT, 0);
           p = grn_text_unesc_tok(ctx, val, p, e, &tok_type);
@@ -1486,6 +1598,9 @@ grn_ctx_qe_exec(grn_ctx *ctx, const char *str, uint32_t str_len)
 exit :
   GRN_OBJ_FIN(ctx, &request_id);
   GRN_OBJ_FIN(ctx, &buf);
+
+  grn_ctx_set_command_version(ctx, command_version);
+
   return expr;
 }
 
@@ -1563,8 +1678,9 @@ grn_ctx_send(grn_ctx *ctx, const char *str, unsigned int str_len, int flags)
         grn_expr_exec(ctx, expr, 0);
       } else {
         if (comment_command_p(str, str_len)) { goto output; };
-        ctx->impl->mime_type = "application/json";
-        ctx->impl->output_type = GRN_CONTENT_JSON;
+        ctx->impl->output.type = GRN_CONTENT_JSON;
+        ctx->impl->output.mime_type = "application/json";
+        ctx->impl->output.is_pretty = GRN_FALSE;
         grn_timeval_now(ctx, &ctx->impl->tv);
         GRN_QUERY_LOG(ctx, GRN_QUERY_LOG_COMMAND,
                       ">", "%.*s", str_len, str);
@@ -1583,8 +1699,8 @@ grn_ctx_send(grn_ctx *ctx, const char *str, unsigned int str_len, int flags)
       }
     output :
       if (!ERRP(ctx, GRN_CRIT)) {
-        if (!(flags & GRN_CTX_QUIET) && ctx->impl->output) {
-          ctx->impl->output(ctx, GRN_CTX_TAIL, ctx->impl->data.ptr);
+        if (!(flags & GRN_CTX_QUIET) && ctx->impl->output.func) {
+          ctx->impl->output.func(ctx, GRN_CTX_TAIL, ctx->impl->output.data.ptr);
         }
       }
       if (expr) { grn_expr_clear_vars(ctx, expr); }
@@ -1610,20 +1726,20 @@ grn_ctx_recv(grn_ctx *ctx, char **str, unsigned int *str_len, int *flags)
   if (ctx->impl) {
     if (ctx->impl->com) {
       grn_com_header header;
-      if (grn_com_recv(ctx, ctx->impl->com, &header, ctx->impl->outbuf)) {
+      if (grn_com_recv(ctx, ctx->impl->com, &header, ctx->impl->output.buf)) {
         *str = NULL;
         *str_len = 0;
         *flags = 0;
       } else {
-        *str = GRN_BULK_HEAD(ctx->impl->outbuf);
-        *str_len = GRN_BULK_VSIZE(ctx->impl->outbuf);
+        *str = GRN_BULK_HEAD(ctx->impl->output.buf);
+        *str_len = GRN_BULK_VSIZE(ctx->impl->output.buf);
         if (header.flags & GRN_CTX_QUIT) {
           ctx->stat = GRN_CTX_QUIT;
           *flags = GRN_CTX_QUIT;
         } else {
           *flags = (header.flags & GRN_CTX_TAIL) ? 0 : GRN_CTX_MORE;
         }
-        ctx->impl->output_type = header.qtype;
+        ctx->impl->output.type = header.qtype;
         ctx->rc = (int16_t)ntohs(header.status);
         ctx->errbuf[0] = '\0';
         ctx->errline = 0;
@@ -1632,11 +1748,11 @@ grn_ctx_recv(grn_ctx *ctx, char **str, unsigned int *str_len, int *flags)
       }
       goto exit;
     } else {
-      grn_obj *buf = ctx->impl->outbuf;
+      grn_obj *buf = ctx->impl->output.buf;
       unsigned int head = 0, tail = GRN_BULK_VSIZE(buf);
       *str = GRN_BULK_HEAD(buf) + head;
       *str_len = tail - head;
-      GRN_BULK_REWIND(ctx->impl->outbuf);
+      GRN_BULK_REWIND(ctx->impl->output.buf);
       goto exit;
     }
   }
@@ -1649,7 +1765,7 @@ void
 grn_ctx_stream_out_func(grn_ctx *ctx, int flags, void *stream)
 {
   if (ctx && ctx->impl) {
-    grn_obj *buf = ctx->impl->outbuf;
+    grn_obj *buf = ctx->impl->output.buf;
     uint32_t size = GRN_BULK_VSIZE(buf);
     if (size) {
       if (fwrite(GRN_BULK_HEAD(buf), 1, size, (FILE *)stream)) {
@@ -1665,8 +1781,8 @@ void
 grn_ctx_recv_handler_set(grn_ctx *ctx, void (*func)(grn_ctx *, int, void *), void *func_arg)
 {
   if (ctx && ctx->impl) {
-    ctx->impl->output = func;
-    ctx->impl->data.ptr = func_arg;
+    ctx->impl->output.func = func;
+    ctx->impl->output.data.ptr = func_arg;
   }
 }
 
@@ -1677,12 +1793,12 @@ grn_ctx_info_get(grn_ctx *ctx, grn_ctx_info *info)
   if (ctx->impl->com) {
     info->fd = ctx->impl->com->fd;
     info->com_status = ctx->impl->com_status;
-    info->outbuf = ctx->impl->outbuf;
+    info->outbuf = ctx->impl->output.buf;
     info->stat = ctx->stat;
   } else {
     info->fd = -1;
     info->com_status = 0;
-    info->outbuf = ctx->impl->outbuf;
+    info->outbuf = ctx->impl->output.buf;
     info->stat = ctx->stat;
   }
   return GRN_SUCCESS;
@@ -2576,6 +2692,12 @@ grn_get_package(void)
   return PACKAGE;
 }
 
+const char *
+grn_get_package_label(void)
+{
+  return PACKAGE_LABEL;
+}
+
 #if defined(HAVE_SIGNAL_H) && !defined(WIN32)
 static int segv_received = 0;
 static void
@@ -2697,79 +2819,123 @@ grn_ctx_output_flush(grn_ctx *ctx, int flags)
   if (flags & GRN_CTX_QUIET) {
     return;
   }
-  if (!ctx->impl->output) {
+  if (!ctx->impl->output.func) {
     return;
   }
-  ctx->impl->output(ctx, 0, ctx->impl->data.ptr);
+  ctx->impl->output.func(ctx, 0, ctx->impl->output.data.ptr);
 }
 
 void
 grn_ctx_output_array_open(grn_ctx *ctx, const char *name, int nelements)
 {
-  grn_output_array_open(ctx, ctx->impl->outbuf, ctx->impl->output_type,
+  grn_output_array_open(ctx,
+                        ctx->impl->output.buf,
+                        ctx->impl->output.type,
                         name, nelements);
 }
 
 void
 grn_ctx_output_array_close(grn_ctx *ctx)
 {
-  grn_output_array_close(ctx, ctx->impl->outbuf, ctx->impl->output_type);
+  grn_output_array_close(ctx,
+                         ctx->impl->output.buf,
+                         ctx->impl->output.type);
 }
 
 void
 grn_ctx_output_map_open(grn_ctx *ctx, const char *name, int nelements)
 {
-  grn_output_map_open(ctx, ctx->impl->outbuf, ctx->impl->output_type,
+  grn_output_map_open(ctx,
+                      ctx->impl->output.buf,
+                      ctx->impl->output.type,
                       name, nelements);
 }
 
 void
 grn_ctx_output_map_close(grn_ctx *ctx)
 {
-  grn_output_map_close(ctx, ctx->impl->outbuf, ctx->impl->output_type);
+  grn_output_map_close(ctx,
+                       ctx->impl->output.buf,
+                       ctx->impl->output.type);
+}
+
+void
+grn_ctx_output_null(grn_ctx *ctx)
+{
+  grn_output_null(ctx,
+                  ctx->impl->output.buf,
+                  ctx->impl->output.type);
 }
 
 void
 grn_ctx_output_int32(grn_ctx *ctx, int value)
 {
-  grn_output_int32(ctx, ctx->impl->outbuf, ctx->impl->output_type, value);
+  grn_output_int32(ctx,
+                   ctx->impl->output.buf,
+                   ctx->impl->output.type,
+                   value);
 }
 
 void
-grn_ctx_output_int64(grn_ctx *ctx, long long int value)
+grn_ctx_output_int64(grn_ctx *ctx, int64_t value)
 {
-  grn_output_int64(ctx, ctx->impl->outbuf, ctx->impl->output_type, value);
+  grn_output_int64(ctx,
+                   ctx->impl->output.buf,
+                   ctx->impl->output.type,
+                   value);
+}
+
+void
+grn_ctx_output_uint64(grn_ctx *ctx, uint64_t value)
+{
+  grn_output_uint64(ctx,
+                    ctx->impl->output.buf,
+                    ctx->impl->output.type,
+                    value);
 }
 
 void
 grn_ctx_output_float(grn_ctx *ctx, double value)
 {
-  grn_output_float(ctx, ctx->impl->outbuf, ctx->impl->output_type, value);
+  grn_output_float(ctx,
+                   ctx->impl->output.buf,
+                   ctx->impl->output.type,
+                   value);
 }
 
 void
 grn_ctx_output_cstr(grn_ctx *ctx, const char *value)
 {
-  grn_output_cstr(ctx, ctx->impl->outbuf, ctx->impl->output_type, value);
+  grn_output_cstr(ctx,
+                  ctx->impl->output.buf,
+                  ctx->impl->output.type,
+                  value);
 }
 
 void
 grn_ctx_output_str(grn_ctx *ctx, const char *value, unsigned int value_len)
 {
-  grn_output_str(ctx, ctx->impl->outbuf, ctx->impl->output_type,
+  grn_output_str(ctx,
+                 ctx->impl->output.buf,
+                 ctx->impl->output.type,
                  value, value_len);
 }
 
 void
 grn_ctx_output_bool(grn_ctx *ctx, grn_bool value)
 {
-  grn_output_bool(ctx, ctx->impl->outbuf, ctx->impl->output_type, value);
+  grn_output_bool(ctx,
+                  ctx->impl->output.buf,
+                  ctx->impl->output.type,
+                  value);
 }
 
 void
 grn_ctx_output_obj(grn_ctx *ctx, grn_obj *value, grn_obj_format *format)
 {
-  grn_output_obj(ctx, ctx->impl->outbuf, ctx->impl->output_type,
+  grn_output_obj(ctx,
+                 ctx->impl->output.buf,
+                 ctx->impl->output.type,
                  value, format);
 }
 
@@ -2778,8 +2944,8 @@ grn_ctx_output_table_columns(grn_ctx *ctx, grn_obj *table,
                              grn_obj_format *format)
 {
   grn_output_table_columns(ctx,
-                           ctx->impl->outbuf,
-                           ctx->impl->output_type,
+                           ctx->impl->output.buf,
+                           ctx->impl->output.type,
                            table,
                            format);
 }
@@ -2789,8 +2955,8 @@ grn_ctx_output_table_records(grn_ctx *ctx, grn_obj *table,
                              grn_obj_format *format)
 {
   grn_output_table_records(ctx,
-                           ctx->impl->outbuf,
-                           ctx->impl->output_type,
+                           ctx->impl->output.buf,
+                           ctx->impl->output.type,
                            table,
                            format);
 }
