@@ -1,5 +1,6 @@
 /* -*- c-basic-offset: 2 -*- */
-/* Copyright(C) 2010-2015 Brazil
+/*
+  Copyright(C) 2010-2016 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -14,8 +15,8 @@
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
-#ifndef GRN_UTIL_H
-#define GRN_UTIL_H
+
+#pragma once
 
 #include "grn.h"
 #include "grn_ctx.h"
@@ -26,10 +27,10 @@ extern "C" {
 
 GRN_API grn_rc grn_normalize_offset_and_limit(grn_ctx *ctx, int size, int *offset, int *limit);
 
-GRN_API const char *grn_win32_base_dir(void);
 GRN_API char *grn_path_separator_to_system(char *dest, char *groonga_path);
 
 int grn_mkstemp(char *path_template);
+grn_bool grn_path_exist(const char *path);
 
 #ifdef WIN32
 GRN_API UINT grn_windows_encoding_to_code_page(grn_encoding encoding);
@@ -38,5 +39,3 @@ GRN_API UINT grn_windows_encoding_to_code_page(grn_encoding encoding);
 #ifdef __cplusplus
 }
 #endif
-
-#endif /* GRN_UTIL_H */
