@@ -62,6 +62,7 @@ typedef struct _grn_io_fileinfo {
 #define IO_HEADER_SIZE 64
 
 static uint32_t grn_io_version_default = GRN_IO_VERSION_DEFAULT;
+static grn_bool grn_io_use_sparse = GRN_FALSE;
 
 inline static grn_rc grn_fileinfo_open(grn_ctx *ctx, fileinfo *fi, const char *path, int flags);
 inline static void grn_fileinfo_init(fileinfo *fis, int nfis);
@@ -104,13 +105,26 @@ inline static grn_rc grn_pwrite(grn_ctx *ctx, fileinfo *fi, void *buf, size_t co
 void
 grn_io_init_from_env(void)
 {
-  char version_env[GRN_ENV_BUFFER_SIZE];
+  {
+    char version_env[GRN_ENV_BUFFER_SIZE];
 
-  grn_getenv("GRN_IO_VERSION",
-             version_env,
-             GRN_ENV_BUFFER_SIZE);
-  if (version_env[0]) {
-    grn_io_version_default = atoi(version_env);
+    grn_getenv("GRN_IO_VERSION",
+               version_env,
+               GRN_ENV_BUFFER_SIZE);
+    if (version_env[0]) {
+      grn_io_version_default = atoi(version_env);
+    }
+  }
+
+  {
+    char use_sparse_env[GRN_ENV_BUFFER_SIZE];
+
+    grn_getenv("GRN_IO_USE_SPARSE",
+               use_sparse_env,
+               GRN_ENV_BUFFER_SIZE);
+    if (use_sparse_env[0] && strcmp(use_sparse_env, "yes") == 0) {
+      grn_io_use_sparse = GRN_TRUE;
+    }
   }
 }
 
@@ -473,7 +487,7 @@ grn_io_detect_type(grn_ctx *ctx, const char *path)
   struct _grn_io_header h;
   uint32_t res = 0;
   int fd;
-  grn_open(fd, path, O_RDWR | GRN_OPEN_FLAG_BINARY);
+  grn_open(fd, path, O_RDONLY | GRN_OPEN_FLAG_BINARY);
   if (fd != -1) {
     struct stat s;
     if (fstat(fd, &s) != -1 && s.st_size >= sizeof(struct _grn_io_header)) {
@@ -487,20 +501,16 @@ grn_io_detect_type(grn_ctx *ctx, const char *path)
               (int)GRN_IO_IDSTR_LEN, GRN_IO_IDSTR);
         }
       } else {
-        SERR(path);
-        GRN_LOG(ctx, GRN_LOG_ERROR,
-                "failed to read enough data for detecting type: <%s>",
-                path);
+        SERR("failed to read enough data for detecting type: <%s>",
+             path);
       }
     } else {
       ERR(GRN_INVALID_FORMAT, "grn_io_detect_type failed");
     }
     grn_close(fd);
   } else {
-    ERRNO_ERR(path);
-    GRN_LOG(ctx, GRN_LOG_ERROR,
-            "failed to open path for detecting type: <%s>",
-            path);
+    ERRNO_ERR("failed to open path for detecting type: <%s>",
+              path);
   }
   return res;
 }
@@ -520,10 +530,8 @@ grn_io_open(grn_ctx *ctx, const char *path, grn_io_mode mode)
     int fd;
     grn_open(fd, path, O_RDWR | GRN_OPEN_FLAG_BINARY);
     if (fd == -1) {
-      ERRNO_ERR(path);
-      GRN_LOG(ctx, GRN_LOG_ERROR,
-              "failed to open path: <%s>",
-              path);
+      ERRNO_ERR("failed to open path: <%s>",
+                path);
       return NULL;
     }
     if (fstat(fd, &s) != -1 && s.st_size >= sizeof(struct _grn_io_header)) {
@@ -718,10 +726,8 @@ grn_io_size(grn_ctx *ctx, grn_io *io, uint64_t *size)
   for (fno = 0; fno < nfiles; fno++) {
     gen_pathname(io->path, buffer, fno);
     if (stat(buffer, &s)) {
-      SERR(buffer);
-      GRN_LOG(ctx, GRN_LOG_ERROR,
-              "failed to stat path to compute size: <%s>",
-              buffer);
+      SERR("failed to stat path to compute size: <%s>",
+           buffer);
     } else {
       tsize += s.st_size;
     }
@@ -735,13 +741,11 @@ grn_io_remove(grn_ctx *ctx, const char *path)
 {
   struct stat s;
   if (stat(path, &s)) {
-    SERR("stat");
+    SERR("failed to stat: <%s>", path);
     return ctx->rc;
   } else if (grn_unlink(path)) {
-    ERRNO_ERR(path);
-    GRN_LOG(ctx, GRN_LOG_ERROR,
-            "failed to remove path: <%s>",
-            path);
+    ERRNO_ERR("failed to remove path: <%s>",
+              path);
     return ctx->rc;
   } else {
     int fno;
@@ -750,10 +754,8 @@ grn_io_remove(grn_ctx *ctx, const char *path)
       gen_pathname(path, buffer, fno);
       if (!stat(buffer, &s)) {
         if (grn_unlink(buffer)) {
-          ERRNO_ERR(buffer);
-          GRN_LOG(ctx, GRN_LOG_ERROR,
-                  "failed to remove path: <%s>",
-                  buffer);
+          ERRNO_ERR("failed to remove path: <%s>",
+                    buffer);
         }
       } else {
         break;
@@ -768,13 +770,11 @@ grn_io_rename(grn_ctx *ctx, const char *old_name, const char *new_name)
 {
   struct stat s;
   if (stat(old_name, &s)) {
-    SERR("stat");
+    SERR("failed to stat path to be renamed: <%s>", old_name);
     return ctx->rc;
   } else if (rename(old_name, new_name)) {
-    SERR(old_name);
-    GRN_LOG(ctx, GRN_LOG_ERROR,
-            "failed to rename path: <%s> -> <%s>",
-            old_name, new_name);
+    SERR("failed to rename path: <%s> -> <%s>",
+         old_name, new_name);
     return ctx->rc;
   } else {
     int fno;
@@ -785,16 +785,12 @@ grn_io_rename(grn_ctx *ctx, const char *old_name, const char *new_name)
       if (!stat(old_buffer, &s)) {
         gen_pathname(new_name, new_buffer, fno);
         if (rename(old_buffer, new_buffer)) {
-          SERR(old_buffer);
-          GRN_LOG(ctx, GRN_LOG_ERROR,
-                  "failed to rename path: <%s> -> <%s>",
-                  old_buffer, new_buffer);
+          SERR("failed to rename path: <%s> -> <%s>",
+               old_buffer, new_buffer);
         }
       } else {
-        SERR("stat");
-        GRN_LOG(ctx, GRN_LOG_ERROR,
-                "failed to stat path to rename: <%s>",
-                old_buffer);
+        SERR("failed to stat path to be renamed: <%s>",
+             old_buffer);
         return ctx->rc;
       }
     }
@@ -1415,20 +1411,16 @@ grn_mmap_v1(grn_ctx *ctx, HANDLE *fmo, fileinfo *fi, off_t offset, size_t length
   /* try to create fmo */
   *fmo = CreateFileMapping(fi->fh, NULL, PAGE_READWRITE, 0, offset + length, NULL);
   if (!*fmo) {
-    SERR("CreateFileMapping");
-    GRN_LOG(ctx, GRN_LOG_ERROR,
-            "CreateFileMapping(%lu + %" GRN_FMT_SIZE ") failed "
-            "<%" GRN_FMT_SIZE ">",
-            (DWORD)offset, length,
-            mmap_size);
+    SERR("CreateFileMapping(%lu + %" GRN_FMT_SIZE ") failed "
+         "<%" GRN_FMT_SIZE ">",
+         (DWORD)offset, length,
+         mmap_size);
     return NULL;
   }
   res = MapViewOfFile(*fmo, FILE_MAP_WRITE, 0, (DWORD)offset, (SIZE_T)length);
   if (!res) {
-    SERR("MapViewOfFile");
-    GRN_LOG(ctx, GRN_LOG_ERROR,
-            "MapViewOfFile(%lu,%" GRN_FMT_SIZE ") failed <%" GRN_FMT_SIZE ">",
-            (DWORD)offset, length, mmap_size);
+    SERR("MapViewOfFile(%lu,%" GRN_FMT_SIZE ") failed <%" GRN_FMT_SIZE ">",
+         (DWORD)offset, length, mmap_size);
     return NULL;
   }
   /* CRITICAL_SECTION_LEAVE(fi->cs); */
@@ -1456,17 +1448,13 @@ grn_munmap_v1(grn_ctx *ctx, HANDLE *fmo, fileinfo *fi,
     if (UnmapViewOfFile(start)) {
       mmap_size -= length;
     } else {
-      SERR("UnmapViewOfFile");
-      GRN_LOG(ctx, GRN_LOG_ERROR,
-              "UnmapViewOfFile(%p,%" GRN_FMT_SIZE ") failed <%" GRN_FMT_SIZE ">",
-              start, length, mmap_size);
+      SERR("UnmapViewOfFile(%p,%" GRN_FMT_SIZE ") failed <%" GRN_FMT_SIZE ">",
+           start, length, mmap_size);
       r = -1;
     }
     if (!CloseHandle(*fmo)) {
-      SERR("CloseHandle");
-      GRN_LOG(ctx, GRN_LOG_ERROR,
-              "CloseHandle(%p,%" GRN_FMT_SIZE ") failed <%" GRN_FMT_SIZE ">",
-              start, length, mmap_size);
+      SERR("CloseHandle(%p,%" GRN_FMT_SIZE ") failed <%" GRN_FMT_SIZE ">",
+           start, length, mmap_size);
     }
     *fmo = NULL;
   } else {
@@ -1553,10 +1541,8 @@ grn_munmap_v0(grn_ctx *ctx, fileinfo *fi, void *start, size_t length)
     mmap_size -= length;
     return 0;
   } else {
-    SERR("UnmapViewOfFile");
-    GRN_LOG(ctx, GRN_LOG_ERROR,
-            "UnmapViewOfFile(%p,%" GRN_FMT_SIZE ") failed <%" GRN_FMT_SIZE ">",
-            start, length, mmap_size);
+    SERR("UnmapViewOfFile(%p,%" GRN_FMT_SIZE ") failed <%" GRN_FMT_SIZE ">",
+         start, length, mmap_size);
     return -1;
   }
 }
@@ -1580,14 +1566,34 @@ grn_fileinfo_open_common(grn_ctx *ctx, fileinfo *fi, const char *path, int flags
                         NULL,
                         dwCreationDisposition, FILE_ATTRIBUTE_NORMAL, 0);
     if (fi->fh == INVALID_HANDLE_VALUE) {
-      SERR("CreateFile");
-      GRN_LOG(ctx, GRN_LOG_ERROR,
-              "CreateFile(<%s>, <%s>) failed",
-              path, flags_description);
+      SERR("CreateFile(<%s>, <%s>) failed",
+           path, flags_description);
       goto exit;
     }
+
+    if (grn_io_use_sparse) {
+      FILE_SET_SPARSE_BUFFER buffer;
+      buffer.SetSparse = TRUE;
+      DWORD returned_bytes;
+      if (!DeviceIoControl(fi->fh,
+                           FSCTL_SET_SPARSE,
+                           &buffer,
+                           sizeof(FILE_SET_SPARSE_BUFFER),
+                           NULL,
+                           0,
+                           &returned_bytes,
+                           NULL)) {
+        GRN_LOG(ctx, GRN_LOG_INFO,
+                "Tried to make file sparse but failed: "
+                "DeviceIoControl(FSCTL_SET_SPARSE): "
+                "<%s>: <%s>",
+                path, grn_current_error_message());
+      }
+    }
+
     goto exit;
   }
+
   if ((flags & O_TRUNC)) {
     CloseHandle(fi->fh);
     /* unable to assign OPEN_ALWAYS and TRUNCATE_EXISTING at once */
@@ -1596,10 +1602,8 @@ grn_fileinfo_open_common(grn_ctx *ctx, fileinfo *fi, const char *path, int flags
                         NULL,
                         TRUNCATE_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
     if (fi->fh == INVALID_HANDLE_VALUE) {
-      SERR("CreateFile");
-      GRN_LOG(ctx, GRN_LOG_ERROR,
-              "CreateFile(<%s>, <O_RDWR|O_TRUNC>) failed",
-              path);
+      SERR("CreateFile(<%s>, <O_RDWR|O_TRUNC>) failed",
+           path);
       goto exit;
     }
     goto exit;
@@ -1610,10 +1614,8 @@ grn_fileinfo_open_common(grn_ctx *ctx, fileinfo *fi, const char *path, int flags
                       NULL,
                       OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
   if (fi->fh == INVALID_HANDLE_VALUE) {
-    SERR("CreateFile");
-    GRN_LOG(ctx, GRN_LOG_ERROR,
-            "CreateFile(<%s>, <O_RDWR>) failed",
-            path);
+    SERR("CreateFile(<%s>, <O_RDWR>) failed",
+         path);
     goto exit;
   }
 
@@ -1738,10 +1740,8 @@ grn_msync(grn_ctx *ctx, void *start, size_t length)
   if (succeeded) {
     return 0;
   } else {
-    SERR("FlushViewOfFile");
-    GRN_LOG(ctx, GRN_LOG_ERROR,
-            "FlushViewOfFile(<%p>, <%" GRN_FMT_SIZE ">) failed",
-            start, length);
+    SERR("FlushViewOfFile(<%p>, <%" GRN_FMT_SIZE ">) failed",
+         start, length);
     return -1;
   }
 }
@@ -1798,17 +1798,13 @@ grn_fileinfo_open(grn_ctx *ctx, fileinfo *fi, const char *path, int flags)
   struct stat st;
   grn_open(fi->fd, path, flags);
   if (fi->fd == -1) {
-    ERRNO_ERR(path);
-    GRN_LOG(ctx, GRN_LOG_ERROR,
-            "failed to open file info path: <%s>",
-            path);
+    ERRNO_ERR("failed to open file info path: <%s>",
+              path);
     return ctx->rc;
   }
   if (fstat(fi->fd, &st) == -1) {
-    ERRNO_ERR(path);
-    GRN_LOG(ctx, GRN_LOG_ERROR,
-            "failed to stat file info path: <%s>",
-            path);
+    ERRNO_ERR("failed to stat file info path: <%s>",
+              path);
     return ctx->rc;
   }
   fi->dev = st.st_dev;
@@ -1913,9 +1909,10 @@ grn_munmap(grn_ctx *ctx, grn_io *io, fileinfo *fi, void *start, size_t length)
   int res;
   res = munmap(start, length);
   if (res) {
-    SERR("munmap");
-    GRN_LOG(ctx, GRN_LOG_ERROR, "munmap(%p,%" GRN_FMT_LLU ") failed <%" GRN_FMT_LLU ">",
-            start, (unsigned long long int)length, (unsigned long long int)mmap_size);
+    SERR("munmap(%p,%" GRN_FMT_LLU ") failed <%" GRN_FMT_LLU ">",
+         start,
+         (unsigned long long int)length,
+         (unsigned long long int)mmap_size);
   } else {
     mmap_size -= length;
   }
