@@ -343,7 +343,7 @@ grn_operator_to_string(grn_operator op)
 grn_bool
 grn_operator_exec_equal(grn_ctx *ctx, grn_obj *x, grn_obj *y)
 {
-  grn_bool r;
+  grn_bool r = GRN_FALSE;
   GRN_API_ENTER;
   DO_EQ(x, y, r);
   GRN_API_RETURN(r);
@@ -352,7 +352,7 @@ grn_operator_exec_equal(grn_ctx *ctx, grn_obj *x, grn_obj *y)
 grn_bool
 grn_operator_exec_not_equal(grn_ctx *ctx, grn_obj *x, grn_obj *y)
 {
-  grn_bool r;
+  grn_bool r = GRN_FALSE;
   GRN_API_ENTER;
   DO_EQ(x, y, r);
   GRN_API_RETURN(!r);
@@ -396,7 +396,7 @@ grn_operator_exec_not_equal(grn_ctx *ctx, grn_obj *x, grn_obj *y)
   }\
 } while (0)
 
-#define DO_COMPARE_SUB(op) do {\
+#define DO_COMPARE_SUB_BUILTIN(op) do {\
   switch (y->header.domain) {\
   case GRN_DB_SHORT_TEXT :\
   case GRN_DB_TEXT :\
@@ -418,6 +418,40 @@ grn_operator_exec_not_equal(grn_ctx *ctx, grn_obj *x, grn_obj *y)
   }\
 } while (0)
 
+#define DO_COMPARE_SUB(op) do {\
+  if (y->header.domain >= GRN_N_RESERVED_TYPES) {\
+    grn_obj *y_table;\
+    y_table = grn_ctx_at(ctx, y->header.domain);\
+    switch (y_table->header.type) {\
+    case GRN_TABLE_HASH_KEY :\
+    case GRN_TABLE_PAT_KEY :\
+    case GRN_TABLE_DAT_KEY :\
+      {\
+        grn_obj y_key;\
+        int length;\
+        GRN_OBJ_INIT(&y_key, GRN_BULK, 0, y_table->header.domain);\
+        length = grn_table_get_key2(ctx, y_table, GRN_RECORD_VALUE(y), &y_key);\
+        if (length > 0) {\
+          grn_obj *y_original = y;\
+          y = &y_key;\
+          DO_COMPARE_SUB_BUILTIN(op);\
+          y = y_original;\
+        } else {\
+          r = GRN_FALSE;\
+        }\
+        GRN_OBJ_FIN(ctx, &y_key);\
+      }\
+      break;\
+    default :\
+      r = GRN_FALSE;\
+      break;\
+    }\
+    grn_obj_unlink(ctx, y_table);\
+  } else {\
+    DO_COMPARE_SUB_BUILTIN(op);\
+  }\
+} while (0)
+
 #define DO_COMPARE_BUILTIN(x,y,r,op) do {\
   switch (x->header.domain) {\
   case GRN_DB_INT8 :\
@@ -544,32 +578,33 @@ grn_operator_exec_not_equal(grn_ctx *ctx, grn_obj *x, grn_obj *y)
 
 #define DO_COMPARE(x, y, r, op) do {\
   if (x->header.domain >= GRN_N_RESERVED_TYPES) {\
-    grn_obj *table;\
-    table = grn_ctx_at(ctx, x->header.domain);\
-    switch (table->header.type) {\
+    grn_obj *x_table;\
+    x_table = grn_ctx_at(ctx, x->header.domain);\
+    switch (x_table->header.type) {\
     case GRN_TABLE_HASH_KEY :\
     case GRN_TABLE_PAT_KEY :\
+    case GRN_TABLE_DAT_KEY :\
       {\
-        grn_obj key;\
+        grn_obj x_key;\
         int length;\
-        GRN_OBJ_INIT(&key, GRN_BULK, 0, table->header.domain);\
-        length = grn_table_get_key2(ctx, table, GRN_RECORD_VALUE(x), &key);\
+        GRN_OBJ_INIT(&x_key, GRN_BULK, 0, x_table->header.domain);\
+        length = grn_table_get_key2(ctx, x_table, GRN_RECORD_VALUE(x), &x_key);\
         if (length > 0) {\
           grn_obj *x_original = x;\
-          x = &key;\
-          DO_COMPARE_BUILTIN((&key), y, r, op);\
+          x = &x_key;\
+          DO_COMPARE_BUILTIN((&x_key), y, r, op);\
           x = x_original;\
         } else {\
           r = GRN_FALSE;\
         }\
-        GRN_OBJ_FIN(ctx, &key);\
+        GRN_OBJ_FIN(ctx, &x_key);\
       }\
       break;\
     default :\
       r = GRN_FALSE;\
       break;\
     }\
-    grn_obj_unlink(ctx, table);\
+    grn_obj_unlink(ctx, x_table);\
   } else {\
     DO_COMPARE_BUILTIN(x, y, r, op);\
   }\
@@ -578,7 +613,7 @@ grn_operator_exec_not_equal(grn_ctx *ctx, grn_obj *x, grn_obj *y)
 grn_bool
 grn_operator_exec_less(grn_ctx *ctx, grn_obj *x, grn_obj *y)
 {
-  grn_bool r;
+  grn_bool r = GRN_FALSE;
   GRN_API_ENTER;
   DO_COMPARE(x, y, r, <);
   GRN_API_RETURN(r);
@@ -587,7 +622,7 @@ grn_operator_exec_less(grn_ctx *ctx, grn_obj *x, grn_obj *y)
 grn_bool
 grn_operator_exec_greater(grn_ctx *ctx, grn_obj *x, grn_obj *y)
 {
-  grn_bool r;
+  grn_bool r = GRN_FALSE;
   GRN_API_ENTER;
   DO_COMPARE(x, y, r, >);
   GRN_API_RETURN(r);
@@ -596,7 +631,7 @@ grn_operator_exec_greater(grn_ctx *ctx, grn_obj *x, grn_obj *y)
 grn_bool
 grn_operator_exec_less_equal(grn_ctx *ctx, grn_obj *x, grn_obj *y)
 {
-  grn_bool r;
+  grn_bool r = GRN_FALSE;
   GRN_API_ENTER;
   DO_COMPARE(x, y, r, <=);
   GRN_API_RETURN(r);
@@ -605,7 +640,7 @@ grn_operator_exec_less_equal(grn_ctx *ctx, grn_obj *x, grn_obj *y)
 grn_bool
 grn_operator_exec_greater_equal(grn_ctx *ctx, grn_obj *x, grn_obj *y)
 {
-  grn_bool r;
+  grn_bool r = GRN_FALSE;
   GRN_API_ENTER;
   DO_COMPARE(x, y, r, >=);
   GRN_API_RETURN(r);
@@ -665,52 +700,20 @@ exec_match_vector_bulk(grn_ctx *ctx, grn_obj *vector, grn_obj *query)
   return matched;
 }
 
-static grn_bool
-string_have_sub_text(grn_ctx *ctx,
-                     const char *text, unsigned int text_len,
-                     const char *sub_text, unsigned int sub_text_len)
-{
-  /* TODO: Use more fast algorithm such as Boyer-Moore algorithm that
-   * is used in snip.c. */
-  const char *text_end = text + text_len;
-  unsigned int sub_text_current = 0;
-
-  for (; text < text_end; text++) {
-    if (text[0] == sub_text[sub_text_current]) {
-      sub_text_current++;
-      if (sub_text_current == sub_text_len) {
-        return GRN_TRUE;
-      }
-    } else {
-      sub_text_current = 0;
-    }
-  }
-
-  return GRN_FALSE;
-}
-
-static grn_bool
-string_have_prefix(grn_ctx *ctx,
-                   const char *target, unsigned int target_len,
-                   const char *prefix, unsigned int prefix_len)
-{
-  return (target_len >= prefix_len &&
-          strncmp(target, prefix, prefix_len) == 0);
-}
-
-static grn_bool
-string_match_regexp(grn_ctx *ctx,
-                    const char *target, unsigned int target_len,
-                    const char *pattern, unsigned int pattern_len)
-{
 #ifdef GRN_SUPPORT_REGEXP
+static OnigRegex
+regexp_compile(grn_ctx *ctx,
+               const char *pattern,
+               unsigned int pattern_len,
+               OnigSyntaxType *syntax)
+{
   OnigRegex regex;
   OnigEncoding onig_encoding;
   int onig_result;
   OnigErrorInfo onig_error_info;
 
   if (ctx->encoding == GRN_ENC_NONE) {
-    return GRN_FALSE;
+    return NULL;
   }
 
   switch (ctx->encoding) {
@@ -730,15 +733,16 @@ string_match_regexp(grn_ctx *ctx,
     onig_encoding = ONIG_ENCODING_KOI8_R;
     break;
   default :
-    return GRN_FALSE;
+    return NULL;
   }
 
   onig_result = onig_new(&regex,
                          pattern,
                          pattern + pattern_len,
-                         ONIG_OPTION_ASCII_RANGE,
+                         ONIG_OPTION_ASCII_RANGE |
+                         ONIG_OPTION_MULTILINE,
                          onig_encoding,
-                         ONIG_SYNTAX_RUBY,
+                         syntax,
                          &onig_error_info);
   if (onig_result != ONIG_NORMAL) {
     char message[ONIG_MAX_ERROR_MESSAGE_LEN];
@@ -748,24 +752,135 @@ string_match_regexp(grn_ctx *ctx,
         "failed to create regular expression object: <%.*s>: %s",
         pattern_len, pattern,
         message);
+    return NULL;
+  }
+
+  return regex;
+}
+
+static grn_bool
+regexp_is_match(grn_ctx *ctx, OnigRegex regex,
+                const char *target, unsigned int target_len)
+{
+  OnigPosition position;
+
+  position = onig_search(regex,
+                         target,
+                         target + target_len,
+                         target,
+                         target + target_len,
+                         NULL,
+                         ONIG_OPTION_NONE);
+  return position != ONIG_MISMATCH;
+}
+#endif /* GRN_SUPPORT_REGEXP */
+
+static grn_bool
+string_have_sub_text(grn_ctx *ctx,
+                     const char *text, unsigned int text_len,
+                     const char *sub_text, unsigned int sub_text_len)
+{
+  if (sub_text_len == 0) {
     return GRN_FALSE;
   }
 
-  {
-    OnigPosition position;
-    position = onig_search(regex,
-                           target,
-                           target + target_len,
-                           target,
-                           target + target_len,
-                           NULL,
-                           ONIG_OPTION_NONE);
-    onig_free(regex);
-    return position != ONIG_MISMATCH;
+  if (sub_text_len > text_len) {
+    return GRN_FALSE;
   }
-#else
+
+#ifdef GRN_SUPPORT_REGEXP
+  {
+    OnigRegex regex;
+    grn_bool matched;
+
+    regex = regexp_compile(ctx, sub_text, sub_text_len, ONIG_SYNTAX_ASIS);
+    if (!regex) {
+      return GRN_FALSE;
+    }
+
+    matched = regexp_is_match(ctx, regex, text, text_len);
+    onig_free(regex);
+    return matched;
+  }
+#else /* GRN_SUPPORT_REGEXP */
+  {
+    const char *text_current = text;
+    const char *text_end = text + text_len;
+    const char *sub_text_current = sub_text;
+    const char *sub_text_end = sub_text + sub_text_len;
+    int sub_text_start_char_len;
+    int sub_text_char_len;
+
+    sub_text_start_char_len = grn_charlen(ctx, sub_text, sub_text_end);
+    if (sub_text_start_char_len == 0) {
+      return GRN_FALSE;
+    }
+    sub_text_char_len = sub_text_start_char_len;
+
+    while (text_current < text_end) {
+      int text_char_len;
+
+      text_char_len = grn_charlen(ctx, text_current, text_end);
+      if (text_char_len == 0) {
+        return GRN_FALSE;
+      }
+
+      if (text_char_len == sub_text_char_len &&
+          memcmp(text_current, sub_text_current, text_char_len) == 0) {
+        sub_text_current += sub_text_char_len;
+        if (sub_text_current == sub_text_end) {
+          return GRN_TRUE;
+        }
+
+        sub_text_char_len = grn_charlen(ctx, sub_text_current, sub_text_end);
+        if (sub_text_char_len == 0) {
+          return GRN_FALSE;
+        }
+      } else {
+        if (sub_text_current != sub_text) {
+          sub_text_current = sub_text;
+          sub_text_char_len = sub_text_start_char_len;
+          continue;
+        }
+      }
+
+      text_current += text_char_len;
+    }
+
+    return GRN_FALSE;
+  }
+#endif /* GRN_SUPPORT_REGEXP */
+}
+
+static grn_bool
+string_have_prefix(grn_ctx *ctx,
+                   const char *target, unsigned int target_len,
+                   const char *prefix, unsigned int prefix_len)
+{
+  return (target_len >= prefix_len &&
+          strncmp(target, prefix, prefix_len) == 0);
+}
+
+static grn_bool
+string_match_regexp(grn_ctx *ctx,
+                    const char *target, unsigned int target_len,
+                    const char *pattern, unsigned int pattern_len)
+{
+#ifdef GRN_SUPPORT_REGEXP
+  OnigRegex regex;
+  grn_bool matched;
+
+  regex = regexp_compile(ctx, pattern, pattern_len, ONIG_SYNTAX_RUBY);
+  if (!regex) {
+    return GRN_FALSE;
+  }
+
+  matched = regexp_is_match(ctx, regex, target, target_len);
+  onig_free(regex);
+  return matched;
+#else /* GRN_SUPPORT_REGEXP */
   return GRN_FALSE;
-#endif
+#endif /* GRN_SUPPORT_REGEXP */
 }
 
 static grn_bool
@@ -778,6 +893,10 @@ exec_text_operator(grn_ctx *ctx,
 {
   grn_bool matched = GRN_FALSE;
 
+  if (target_len == 0 || query_len == 0) {
+    return GRN_FALSE;
+  }
+
   switch (op) {
   case GRN_OP_MATCH :
     matched = string_have_sub_text(ctx, target, target_len, query, query_len);
@@ -817,23 +936,24 @@ exec_text_operator_raw_text_raw_text(grn_ctx *ctx,
     return GRN_FALSE;
   }
 
-  if (op == GRN_OP_REGEXP) {
-    return exec_text_operator(ctx, op,
-                              target, target_len,
-                              query, query_len);
-  }
-
   normalizer = grn_ctx_get(ctx, GRN_NORMALIZER_AUTO_NAME, -1);
   norm_target = grn_string_open(ctx, target, target_len, normalizer, 0);
-  norm_query  = grn_string_open(ctx, query,  query_len,  normalizer, 0);
   grn_string_get_normalized(ctx, norm_target,
                             &norm_target_raw,
                             &norm_target_raw_length_in_bytes,
                             NULL);
-  grn_string_get_normalized(ctx, norm_query,
-                            &norm_query_raw,
-                            &norm_query_raw_length_in_bytes,
-                            NULL);
+
+  if (op == GRN_OP_REGEXP) {
+    norm_query = NULL;
+    norm_query_raw = query;
+    norm_query_raw_length_in_bytes = query_len;
+  } else {
+    norm_query = grn_string_open(ctx, query,  query_len,  normalizer, 0);
+    grn_string_get_normalized(ctx, norm_query,
+                              &norm_query_raw,
+                              &norm_query_raw_length_in_bytes,
+                              NULL);
+  }
 
   matched = exec_text_operator(ctx, op,
                                norm_target_raw,
@@ -842,7 +962,9 @@ exec_text_operator_raw_text_raw_text(grn_ctx *ctx,
                                norm_query_raw_length_in_bytes);
 
   grn_obj_close(ctx, norm_target);
-  grn_obj_close(ctx, norm_query);
+  if (norm_query) {
+    grn_obj_close(ctx, norm_query);
+  }
   grn_obj_unlink(ctx, normalizer);
 
   return matched;
@@ -870,26 +992,35 @@ exec_text_operator_record_text(grn_ctx *ctx,
   record_key_len = grn_table_get_key(ctx, table, GRN_RECORD_VALUE(record),
                                      record_key, GRN_TABLE_MAX_KEY_SIZE);
   grn_table_get_info(ctx, table, NULL, NULL, NULL, &normalizer, NULL);
-  if (normalizer && (op != GRN_OP_REGEXP)) {
+  if (normalizer) {
     grn_obj *norm_query;
     const char *norm_query_raw;
     unsigned int norm_query_raw_length_in_bytes;
-    norm_query = grn_string_open(ctx,
-                                 GRN_TEXT_VALUE(query),
-                                 GRN_TEXT_LEN(query),
-                                 normalizer,
-                                 0);
-    grn_string_get_normalized(ctx, norm_query,
-                              &norm_query_raw,
-                              &norm_query_raw_length_in_bytes,
-                              NULL);
+
+    if (op == GRN_OP_REGEXP) {
+      norm_query = NULL;
+      norm_query_raw = GRN_TEXT_VALUE(query);
+      norm_query_raw_length_in_bytes = GRN_TEXT_LEN(query);
+    } else {
+      norm_query = grn_string_open(ctx,
+                                   GRN_TEXT_VALUE(query),
+                                   GRN_TEXT_LEN(query),
+                                   normalizer,
+                                   0);
+      grn_string_get_normalized(ctx, norm_query,
+                                &norm_query_raw,
+                                &norm_query_raw_length_in_bytes,
+                                NULL);
+    }
     matched = exec_text_operator(ctx,
                                  op,
                                  record_key,
                                  record_key_len,
                                  norm_query_raw,
                                  norm_query_raw_length_in_bytes);
-    grn_obj_close(ctx, norm_query);
+    if (norm_query) {
+      grn_obj_close(ctx, norm_query);
+    }
   } else {
     matched = exec_text_operator_raw_text_raw_text(ctx,
                                                    op,
@@ -982,11 +1113,171 @@ grn_operator_exec_prefix(grn_ctx *ctx, grn_obj *target, grn_obj *prefix)
   GRN_API_RETURN(matched);
 }
 
+static grn_bool
+exec_regexp_uvector_bulk(grn_ctx *ctx, grn_obj *uvector, grn_obj *pattern)
+{
+#ifdef GRN_SUPPORT_REGEXP
+  grn_bool matched = GRN_FALSE;
+  unsigned int i, size;
+  OnigRegex regex;
+  grn_obj *domain;
+  grn_obj *normalizer;
+  grn_obj *normalizer_auto = NULL;
+
+  size = grn_uvector_size(ctx, uvector);
+  if (size == 0) {
+    return GRN_FALSE;
+  }
+
+  regex = regexp_compile(ctx,
+                         GRN_TEXT_VALUE(pattern),
+                         GRN_TEXT_LEN(pattern),
+                         ONIG_SYNTAX_RUBY);
+  if (!regex) {
+    return GRN_FALSE;
+  }
+
+  domain = grn_ctx_at(ctx, uvector->header.domain);
+  if (!domain) {
+    onig_free(regex);
+    return GRN_FALSE;
+  }
+
+  grn_table_get_info(ctx, domain, NULL, NULL, NULL, &normalizer, NULL);
+  if (!normalizer) {
+    normalizer_auto = grn_ctx_get(ctx, GRN_NORMALIZER_AUTO_NAME, -1);
+  }
+
+  for (i = 0; i < size; i++) {
+    grn_id record_id;
+    char key[GRN_TABLE_MAX_KEY_SIZE];
+    int key_size;
+
+    record_id = grn_uvector_get_element(ctx, uvector, i, NULL);
+    key_size = grn_table_get_key(ctx, domain, record_id,
+                                 key, GRN_TABLE_MAX_KEY_SIZE);
+    if (key_size == 0) {
+      continue;
+    }
+
+    if (normalizer) {
+      matched = regexp_is_match(ctx, regex, key, key_size);
+    } else {
+      grn_obj *norm_key;
+      const char *norm_key_raw;
+      unsigned int norm_key_raw_length_in_bytes;
+
+      norm_key = grn_string_open(ctx, key, key_size, normalizer_auto, 0);
+      grn_string_get_normalized(ctx, norm_key,
+                                &norm_key_raw,
+                                &norm_key_raw_length_in_bytes,
+                                NULL);
+      matched = regexp_is_match(ctx, regex,
+                                norm_key_raw,
+                                norm_key_raw_length_in_bytes);
+      grn_obj_unlink(ctx, norm_key);
+    }
+
+    if (matched) {
+      break;
+    }
+  }
+
+  if (normalizer_auto) {
+    grn_obj_unlink(ctx, normalizer_auto);
+  }
+
+  grn_obj_unlink(ctx, domain);
+
+  onig_free(regex);
+
+  return matched;
+#else /* GRN_SUPPORT_REGEXP */
+  return GRN_FALSE;
+#endif /* GRN_SUPPORT_REGEXP */
+}
+
+static grn_bool
+exec_regexp_vector_bulk(grn_ctx *ctx, grn_obj *vector, grn_obj *pattern)
+{
+#ifdef GRN_SUPPORT_REGEXP
+  grn_obj *normalizer = NULL;
+  grn_bool matched = GRN_FALSE;
+  unsigned int i, size;
+  OnigRegex regex;
+
+  size = grn_vector_size(ctx, vector);
+  if (size == 0) {
+    return GRN_FALSE;
+  }
+
+  regex = regexp_compile(ctx,
+                         GRN_TEXT_VALUE(pattern),
+                         GRN_TEXT_LEN(pattern),
+                         ONIG_SYNTAX_RUBY);
+  if (!regex) {
+    return GRN_FALSE;
+  }
+
+  normalizer = grn_ctx_get(ctx, GRN_NORMALIZER_AUTO_NAME, -1);
+  for (i = 0; i < size; i++) {
+    const char *content;
+    unsigned int content_size;
+    grn_id domain_id;
+    grn_obj *norm_content;
+    const char *norm_content_raw;
+    unsigned int norm_content_raw_length_in_bytes;
+
+    content_size = grn_vector_get_element(ctx, vector, i,
+                                          &content, NULL, &domain_id);
+    if (content_size == 0) {
+      continue;
+    }
+
+    norm_content = grn_string_open(ctx, content, content_size, normalizer, 0);
+    grn_string_get_normalized(ctx, norm_content,
+                              &norm_content_raw,
+                              &norm_content_raw_length_in_bytes,
+                              NULL);
+
+    matched = regexp_is_match(ctx, regex,
+                              norm_content_raw,
+                              norm_content_raw_length_in_bytes);
+
+    grn_obj_unlink(ctx, norm_content);
+
+    if (matched) {
+      break;
+    }
+  }
+  grn_obj_unlink(ctx, normalizer);
+
+  onig_free(regex);
+
+  return matched;
+#else /* GRN_SUPPORT_REGEXP */
+  return GRN_FALSE;
+#endif /* GRN_SUPPORT_REGEXP */
+}
+
 grn_bool
 grn_operator_exec_regexp(grn_ctx *ctx, grn_obj *target, grn_obj *pattern)
 {
-  grn_bool matched;
+  grn_bool matched = GRN_FALSE;
   GRN_API_ENTER;
-  matched = exec_text_operator_bulk_bulk(ctx, GRN_OP_REGEXP, target, pattern);
+  switch (target->header.type) {
+  case GRN_UVECTOR :
+    matched = exec_regexp_uvector_bulk(ctx, target, pattern);
+    break;
+  case GRN_VECTOR :
+    matched = exec_regexp_vector_bulk(ctx, target, pattern);
+    break;
+  case GRN_BULK :
+    matched = exec_text_operator_bulk_bulk(ctx, GRN_OP_REGEXP, target, pattern);
+    break;
+  default :
+    matched = GRN_FALSE;
+    break;
+  }
   GRN_API_RETURN(matched);
 }
