@@ -22,6 +22,17 @@
 #include <stdio.h>
 #include <string.h>
 
+#include <sys/stat.h>
+#ifdef HAVE_DIRENT_H
+# include <dirent.h>
+#endif /* HAVE_DIRENT_H */
+
+#ifndef S_ISREG
+# ifdef _S_IFREG
+#  define S_ISREG(mode) (mode & _S_IFREG)
+# endif /* _S_IFREG */
+#endif /* !S_ISREG */
+
 #include "grn_db.h"
 #include "grn_plugin.h"
 #include "grn_ctx_impl.h"
@@ -34,10 +45,6 @@
 static grn_hash *grn_plugins = NULL;
 static grn_critical_section grn_plugins_lock;
 
-#ifdef GRN_WITH_MRUBY
-static const char *grn_plugin_mrb_suffix = ".rb";
-#endif /* GRN_WITH_MRUBY */
-
 #ifdef HAVE_DLFCN_H
 #  include <dlfcn.h>
 #  define grn_dl_open(filename)      dlopen(filename, RTLD_LAZY | RTLD_LOCAL)
@@ -57,6 +64,18 @@ static const char *grn_plugin_mrb_suffix = ".rb";
 #  define grn_dl_clear_error()
 #endif
 
+#define GRN_PLUGIN_KEY_SIZE(filename) (strlen((filename)) + 1)
+
+static char grn_plugins_dir[GRN_ENV_BUFFER_SIZE];
+
+void
+grn_plugin_init_from_env(void)
+{
+  grn_getenv("GRN_PLUGINS_DIR",
+             grn_plugins_dir,
+             GRN_ENV_BUFFER_SIZE);
+}
+
 static int
 compute_name_size(const char *name, int name_size)
 {
@@ -77,7 +96,8 @@ grn_plugin_reference(grn_ctx *ctx, const char *filename)
   grn_plugin **plugin = NULL;
 
   CRITICAL_SECTION_ENTER(grn_plugins_lock);
-  id = grn_hash_get(&grn_gctx, grn_plugins, filename, strlen(filename),
+  id = grn_hash_get(&grn_gctx, grn_plugins,
+                    filename, GRN_PLUGIN_KEY_SIZE(filename),
                     (void **)&plugin);
   if (plugin) {
     (*plugin)->refcount++;
@@ -91,7 +111,8 @@ const char *
 grn_plugin_path(grn_ctx *ctx, grn_id id)
 {
   const char *path;
-  uint32_t key_size;
+  grn_plugin *plugin;
+  int value_size;
   const char *system_plugins_dir;
   size_t system_plugins_dir_size;
 
@@ -100,13 +121,14 @@ grn_plugin_path(grn_ctx *ctx, grn_id id)
   }
 
   CRITICAL_SECTION_ENTER(grn_plugins_lock);
-  path = _grn_hash_key(&grn_gctx, grn_plugins, id, &key_size);
+  value_size = grn_hash_get_value(&grn_gctx, grn_plugins, id, &plugin);
   CRITICAL_SECTION_LEAVE(grn_plugins_lock);
 
-  if (!path) {
+  if (!plugin) {
     return NULL;
   }
 
+  path = plugin->path;
   system_plugins_dir = grn_plugin_get_system_plugins_dir();
   system_plugins_dir_size = strlen(system_plugins_dir);
   if (strncmp(system_plugins_dir, path, system_plugins_dir_size) == 0) {
@@ -207,7 +229,7 @@ grn_plugin_initialize(grn_ctx *ctx, grn_plugin *plugin,
   if (!plugin->type ## _func) {                                         \
     const char *label;                                                  \
     label = grn_dl_sym_error_label();                                   \
-    SERR(label);                                                        \
+    SERR("%s", label);                                                  \
   }                                                                     \
 } while (0)
 
@@ -243,6 +265,11 @@ grn_plugin_open_mrb(grn_ctx *ctx, const char *filename, size_t filename_size)
   grn_id id = GRN_ID_NIL;
   grn_plugin **plugin = NULL;
 
+  if (!ctx->impl->mrb.state) {
+    ERR(GRN_FUNCTION_NOT_IMPLEMENTED, "mruby support isn't enabled");
+    return GRN_ID_NIL;
+  }
+
   id = grn_hash_add(&grn_gctx, grn_plugins, filename, filename_size,
                     (void **)&plugin, NULL);
   if (!id) {
@@ -255,6 +282,7 @@ grn_plugin_open_mrb(grn_ctx *ctx, const char *filename, size_t filename_size)
     return GRN_ID_NIL;
   }
 
+  grn_memcpy((*plugin)->path, filename, filename_size);
   (*plugin)->dl = NULL;
   (*plugin)->init_func = NULL;
   (*plugin)->register_func = NULL;
@@ -273,7 +301,7 @@ grn_plugin_open(grn_ctx *ctx, const char *filename)
   grn_plugin **plugin = NULL;
   size_t filename_size;
 
-  filename_size = strlen(filename);
+  filename_size = GRN_PLUGIN_KEY_SIZE(filename);
 
   CRITICAL_SECTION_ENTER(grn_plugins_lock);
   if ((id = grn_hash_get(&grn_gctx, grn_plugins, filename, filename_size,
@@ -283,11 +311,15 @@ grn_plugin_open(grn_ctx *ctx, const char *filename)
   }
 
 #ifdef GRN_WITH_MRUBY
-  if (filename_size > strlen(grn_plugin_mrb_suffix) &&
-      strcmp(filename + (filename_size - strlen(grn_plugin_mrb_suffix)),
-             grn_plugin_mrb_suffix) == 0) {
-    id = grn_plugin_open_mrb(ctx, filename, filename_size);
-    goto exit;
+  {
+    const char *mrb_suffix;
+    mrb_suffix = grn_plugin_get_ruby_suffix();
+    if (filename_size > strlen(mrb_suffix) &&
+      strcmp(filename + (strlen(filename) - strlen(mrb_suffix)),
+             mrb_suffix) == 0) {
+      id = grn_plugin_open_mrb(ctx, filename, filename_size);
+      goto exit;
+    }
   }
 #endif /* GRN_WITH_MRUBY */
 
@@ -296,6 +328,7 @@ grn_plugin_open(grn_ctx *ctx, const char *filename)
                            (void **)&plugin, NULL))) {
       *plugin = GRN_GMALLOCN(grn_plugin, 1);
       if (*plugin) {
+        grn_memcpy((*plugin)->path, filename, filename_size);
         if (grn_plugin_initialize(ctx, *plugin, dl, id, filename)) {
           GRN_GFREE(*plugin);
           *plugin = NULL;
@@ -310,7 +343,7 @@ grn_plugin_open(grn_ctx *ctx, const char *filename)
         } else {
           const char *label;
           label = grn_dl_close_error_label();
-          SERR(label);
+          SERR("%s", label);
         }
         id = GRN_ID_NIL;
       } else {
@@ -320,13 +353,13 @@ grn_plugin_open(grn_ctx *ctx, const char *filename)
       if (!grn_dl_close(dl)) {
         const char *label;
         label = grn_dl_close_error_label();
-        SERR(label);
+        SERR("%s", label);
       }
     }
   } else {
     const char *label;
     label = grn_dl_open_error_label();
-    SERR(label);
+    SERR("%s", label);
   }
 
 exit:
@@ -359,7 +392,7 @@ grn_plugin_close(grn_ctx *ctx, grn_id id)
     if (!grn_dl_close(plugin->dl)) {
       const char *label;
       label = grn_dl_close_error_label();
-      SERR(label);
+      SERR("%s", label);
     }
   }
   GRN_GFREE(plugin);
@@ -390,7 +423,7 @@ grn_plugin_sym(grn_ctx *ctx, grn_id id, const char *symbol)
   if (!(func = grn_dl_sym(plugin->dl, symbol))) {
     const char *label;
     label = grn_dl_sym_error_label();
-    SERR(label);
+    SERR("%s", label);
   }
 
 exit:
@@ -428,6 +461,12 @@ grn_plugin_get_suffix(void)
   return GRN_PLUGIN_SUFFIX;
 }
 
+const char *
+grn_plugin_get_ruby_suffix(void)
+{
+  return ".rb";
+}
+
 grn_rc
 grn_plugin_register_by_path(grn_ctx *ctx, const char *path)
 {
@@ -453,57 +492,73 @@ grn_plugin_register_by_path(grn_ctx *ctx, const char *path)
 }
 
 #ifdef WIN32
-static char *win32_plugins_dir = NULL;
-static char win32_plugins_dir_buffer[PATH_MAX];
-const char *
-grn_plugin_get_system_plugins_dir(void)
+static char *windows_plugins_dir = NULL;
+static char windows_plugins_dir_buffer[PATH_MAX];
+static const char *
+grn_plugin_get_default_system_plugins_dir(void)
 {
-  if (!win32_plugins_dir) {
+  if (!windows_plugins_dir) {
     const char *base_dir;
     const char *relative_path = GRN_RELATIVE_PLUGINS_DIR;
     size_t base_dir_length;
 
-    base_dir = grn_win32_base_dir();
+    base_dir = grn_windows_base_dir();
     base_dir_length = strlen(base_dir);
-    strcpy(win32_plugins_dir_buffer, base_dir);
-    strcat(win32_plugins_dir_buffer, "/");
-    strcat(win32_plugins_dir_buffer, relative_path);
-    win32_plugins_dir = win32_plugins_dir_buffer;
+    grn_strcpy(windows_plugins_dir_buffer, PATH_MAX, base_dir);
+    grn_strcat(windows_plugins_dir_buffer, PATH_MAX, "/");
+    grn_strcat(windows_plugins_dir_buffer, PATH_MAX, relative_path);
+    windows_plugins_dir = windows_plugins_dir_buffer;
   }
-  return win32_plugins_dir;
+  return windows_plugins_dir;
 }
 
 #else /* WIN32 */
-const char *
-grn_plugin_get_system_plugins_dir(void)
+static const char *
+grn_plugin_get_default_system_plugins_dir(void)
 {
   return GRN_PLUGINS_DIR;
 }
 #endif /* WIN32 */
 
+const char *
+grn_plugin_get_system_plugins_dir(void)
+{
+  if (grn_plugins_dir[0]) {
+    return grn_plugins_dir;
+  } else {
+    return grn_plugin_get_default_system_plugins_dir();
+  }
+}
+
 static char *
 grn_plugin_find_path_raw(grn_ctx *ctx, const char *path)
 {
-  FILE *plugin_file;
+  struct stat path_stat;
 
-  plugin_file = fopen(path, "r");
+  if (stat(path, &path_stat) != 0) {
+    return NULL;
+  }
 
-  if (!plugin_file) {
+  if (!S_ISREG(path_stat.st_mode)) {
     return NULL;
   }
 
-  fclose(plugin_file);
   return GRN_STRDUP(path);
 }
 
-#if GRN_WITH_MRUBY
+#ifdef GRN_WITH_MRUBY
 static char *
 grn_plugin_find_path_mrb(grn_ctx *ctx, const char *path, size_t path_len)
 {
   char mrb_path[PATH_MAX];
-  const char *mrb_suffix = grn_plugin_mrb_suffix;
+  const char *mrb_suffix;
   size_t mrb_path_len;
 
+  mrb_suffix = grn_plugin_get_ruby_suffix();
+  if (!ctx->impl->mrb.state) {
+    return NULL;
+  }
+
   mrb_path_len = path_len + strlen(mrb_suffix);
   if (mrb_path_len >= PATH_MAX) {
     ERR(GRN_FILENAME_TOO_LONG,
@@ -512,8 +567,8 @@ grn_plugin_find_path_mrb(grn_ctx *ctx, const char *path, size_t path_len)
     return NULL;
   }
 
-  strcpy(mrb_path, path);
-  strcat(mrb_path, mrb_suffix);
+  grn_strcpy(mrb_path, PATH_MAX, path);
+  grn_strcat(mrb_path, PATH_MAX, mrb_suffix);
   return grn_plugin_find_path_raw(ctx, mrb_path);
 }
 #else /* GRN_WITH_MRUBY */
@@ -540,8 +595,8 @@ grn_plugin_find_path_so(grn_ctx *ctx, const char *path, size_t path_len)
     return NULL;
   }
 
-  strcpy(so_path, path);
-  strcat(so_path, so_suffix);
+  grn_strcpy(so_path, PATH_MAX, path);
+  grn_strcat(so_path, PATH_MAX, so_suffix);
   return grn_plugin_find_path_raw(ctx, so_path);
 }
 
@@ -573,10 +628,10 @@ grn_plugin_find_path_libs_so(grn_ctx *ctx, const char *path, size_t path_len)
   }
 
   libs_so_path[0] = '\0';
-  strncat(libs_so_path, path, base_name - path);
-  strcat(libs_so_path, libs_path);
-  strcat(libs_so_path, base_name);
-  strcat(libs_so_path, so_suffix);
+  grn_strncat(libs_so_path, PATH_MAX, path, base_name - path);
+  grn_strcat(libs_so_path, PATH_MAX, libs_path);
+  grn_strcat(libs_so_path, PATH_MAX, base_name);
+  grn_strcat(libs_so_path, PATH_MAX, so_suffix);
   return grn_plugin_find_path_raw(ctx, libs_so_path);
 }
 
@@ -594,15 +649,12 @@ grn_plugin_find_path(grn_ctx *ctx, const char *name)
   if (name[0] == '/') {
     path[0] = '\0';
   } else {
-    plugins_dir = getenv("GRN_PLUGINS_DIR");
-    if (!plugins_dir) {
-      plugins_dir = grn_plugin_get_system_plugins_dir();
-    }
-    strcpy(path, plugins_dir);
+    plugins_dir = grn_plugin_get_system_plugins_dir();
+    grn_strcpy(path, PATH_MAX, plugins_dir);
 
     dir_last_char = plugins_dir[strlen(path) - 1];
     if (dir_last_char != '/') {
-      strcat(path, "/");
+      grn_strcat(path, PATH_MAX, "/");
     }
   }
 
@@ -615,7 +667,7 @@ grn_plugin_find_path(grn_ctx *ctx, const char *name)
         path, name);
     goto exit;
   }
-  strcat(path, name);
+  grn_strcat(path, PATH_MAX, name);
 
   found_path = grn_plugin_find_path_raw(ctx, path);
   if (found_path) {
@@ -651,6 +703,30 @@ exit :
   GRN_API_RETURN(found_path);
 }
 
+static void
+grn_plugin_set_name_resolve_error(grn_ctx *ctx, const char *name,
+                                  const char *tag)
+{
+  const char *prefix, *prefix_separator, *suffix;
+
+  if (name[0] == '/') {
+    prefix = "";
+    prefix_separator = "";
+    suffix = "";
+  } else {
+    prefix = grn_plugin_get_system_plugins_dir();
+    if (prefix[strlen(prefix) - 1] != '/') {
+      prefix_separator = "/";
+    } else {
+      prefix_separator = "";
+    }
+    suffix = grn_plugin_get_suffix();
+  }
+  ERR(GRN_NO_SUCH_FILE_OR_DIRECTORY,
+      "%s cannot find plugin file: <%s%s%s%s>",
+      tag, prefix, prefix_separator, name, suffix);
+}
+
 grn_rc
 grn_plugin_register(grn_ctx *ctx, const char *name)
 {
@@ -664,26 +740,86 @@ grn_plugin_register(grn_ctx *ctx, const char *name)
     GRN_FREE(path);
   } else {
     if (ctx->rc == GRN_SUCCESS) {
-      const char *prefix, *prefix_separator, *suffix;
-      if (name[0] == '/') {
-        prefix = "";
-        prefix_separator = "";
-        suffix = "";
-      } else {
-        prefix = getenv("GRN_PLUGINS_DIR");
-        if (!prefix) {
-          prefix = grn_plugin_get_system_plugins_dir();
-        }
-        if (prefix[strlen(prefix) - 1] != '/') {
-          prefix_separator = "/";
-        } else {
-          prefix_separator = "";
-        }
-        suffix = grn_plugin_get_suffix();
+      grn_plugin_set_name_resolve_error(ctx, name, "[plugin][register]");
+    }
+    rc = ctx->rc;
+  }
+  GRN_API_RETURN(rc);
+}
+
+grn_rc
+grn_plugin_unregister_by_path(grn_ctx *ctx, const char *path)
+{
+  grn_obj *db;
+  grn_id plugin_id;
+
+  if (!ctx || !ctx->impl) {
+    ERR(GRN_INVALID_ARGUMENT, "[plugin][unregister] ctx isn't initialized");
+    return ctx->rc;
+  }
+
+  db = ctx->impl->db;
+  if (!db) {
+    ERR(GRN_INVALID_ARGUMENT, "[plugin][unregister] DB isn't initialized");
+    return ctx->rc;
+  }
+
+  GRN_API_ENTER;
+
+  CRITICAL_SECTION_ENTER(grn_plugins_lock);
+  plugin_id = grn_hash_get(&grn_gctx, grn_plugins,
+                           path, GRN_PLUGIN_KEY_SIZE(path),
+                           NULL);
+  CRITICAL_SECTION_LEAVE(grn_plugins_lock);
+
+  if (plugin_id == GRN_ID_NIL) {
+    GRN_API_RETURN(ctx->rc);
+  }
+
+  {
+    grn_table_cursor *cursor;
+    grn_id id;
+
+    cursor = grn_table_cursor_open(ctx, db,
+                                   NULL, 0,
+                                   NULL, 0,
+                                   0, -1, GRN_CURSOR_BY_ID);
+    if (!cursor) {
+      GRN_API_RETURN(ctx->rc);
+    }
+
+    while ((id = grn_table_cursor_next(ctx, cursor))) {
+      grn_obj *obj;
+      obj = grn_ctx_at(ctx, id);
+      if (!obj) {
+        continue;
       }
-      ERR(GRN_NO_SUCH_FILE_OR_DIRECTORY,
-          "cannot find plugin file: <%s%s%s%s>",
-          prefix, prefix_separator, name, suffix);
+      if (obj->header.type == GRN_PROC && DB_OBJ(obj)->range == plugin_id) {
+        grn_obj_remove(ctx, obj);
+      } else {
+        grn_obj_unlink(ctx, obj);
+      }
+    }
+    grn_table_cursor_close(ctx, cursor);
+  }
+
+  GRN_API_RETURN(ctx->rc);
+}
+
+grn_rc
+grn_plugin_unregister(grn_ctx *ctx, const char *name)
+{
+  grn_rc rc;
+  char *path;
+
+  GRN_API_ENTER;
+  path = grn_plugin_find_path(ctx, name);
+  if (path) {
+    rc = grn_plugin_unregister_by_path(ctx, path);
+    GRN_FREE(path);
+  } else {
+    if (ctx->rc == GRN_SUCCESS) {
+      grn_plugin_set_name_resolve_error(ctx, name, "[plugin][unregister]");
     }
     rc = ctx->rc;
   }
@@ -741,6 +877,121 @@ grn_plugin_ensure_registered(grn_ctx *ctx, grn_obj *proc)
 #endif /* GRN_WITH_MRUBY */
 }
 
+grn_rc
+grn_plugin_get_names(grn_ctx *ctx, grn_obj *names)
+{
+  grn_obj *db;
+  grn_table_cursor *cursor;
+  grn_id id;
+  grn_hash *processed_paths;
+  const char *system_plugins_dir;
+  const char *native_plugin_suffix;
+  const char *ruby_plugin_suffix;
+
+  GRN_API_ENTER;
+
+  if (ctx->rc) {
+    GRN_API_RETURN(ctx->rc);
+  }
+
+  db = ctx->impl->db;
+  cursor = grn_table_cursor_open(ctx, db, NULL, 0, NULL, 0, 0, -1,
+                                 GRN_CURSOR_BY_ID);
+  if (!cursor) {
+    GRN_API_RETURN(ctx->rc);
+  }
+
+  processed_paths = grn_hash_create(ctx, NULL, GRN_TABLE_MAX_KEY_SIZE, 0,
+                                    GRN_OBJ_TABLE_HASH_KEY |
+                                    GRN_OBJ_KEY_VAR_SIZE);
+  if (!processed_paths) {
+    grn_table_cursor_close(ctx, cursor);
+    GRN_API_RETURN(ctx->rc);
+  }
+
+  system_plugins_dir = grn_plugin_get_system_plugins_dir();
+  native_plugin_suffix = grn_plugin_get_suffix();
+  ruby_plugin_suffix = grn_plugin_get_ruby_suffix();
+  while ((id = grn_table_cursor_next(ctx, cursor)) != GRN_ID_NIL) {
+    grn_obj *object;
+    const char *path;
+    grn_id processed_path_id;
+
+    object = grn_ctx_at(ctx, id);
+    if (!object) {
+      ERRCLR(ctx);
+      continue;
+    }
+
+    if (!grn_obj_is_proc(ctx, object)) {
+      grn_obj_unlink(ctx, object);
+      continue;
+    }
+
+    if (grn_obj_is_builtin(ctx, object)) {
+      grn_obj_unlink(ctx, object);
+      continue;
+    }
+
+    path = grn_obj_path(ctx, object);
+    if (!path) {
+      grn_obj_unlink(ctx, object);
+      continue;
+    }
+
+    processed_path_id = grn_hash_get(ctx, processed_paths,
+                                     path, strlen(path),
+                                     NULL);
+    if (processed_path_id != GRN_ID_NIL) {
+      grn_obj_unlink(ctx, object);
+      continue;
+    }
+
+    grn_hash_add(ctx, processed_paths,
+                 path, strlen(path),
+                 NULL, NULL);
+
+    {
+      const char *relative_path;
+      const char *libs_path = "/.libs/";
+      const char *start_libs;
+      char name[PATH_MAX];
+
+      name[0] = '\0';
+      if (strncmp(path, system_plugins_dir, strlen(system_plugins_dir)) == 0) {
+        relative_path = path + strlen(system_plugins_dir);
+      } else {
+        relative_path = path;
+      }
+      start_libs = strstr(relative_path, libs_path);
+      if (start_libs) {
+        grn_strncat(name, PATH_MAX, relative_path, start_libs - relative_path);
+        grn_strcat(name, PATH_MAX, "/");
+        grn_strcat(name, PATH_MAX, start_libs + strlen(libs_path));
+      } else {
+        grn_strcat(name, PATH_MAX, relative_path);
+      }
+      if (strlen(name) > strlen(native_plugin_suffix) &&
+          strcmp(name + strlen(name) - strlen(native_plugin_suffix),
+                 native_plugin_suffix) == 0) {
+        name[strlen(name) - strlen(native_plugin_suffix)] = '\0';
+      } else if (strlen(name) > strlen(ruby_plugin_suffix) &&
+                 strcmp(name + strlen(name) - strlen(ruby_plugin_suffix),
+                        ruby_plugin_suffix) == 0) {
+        name[strlen(name) - strlen(ruby_plugin_suffix)] = '\0';
+      }
+      grn_vector_add_element(ctx, names,
+                             name, strlen(name),
+                             0, GRN_DB_TEXT);
+    }
+  }
+  grn_table_cursor_close(ctx, cursor);
+
+  grn_hash_close(ctx, processed_paths);
+
+  GRN_API_RETURN(ctx->rc);
+}
+
 void *
 grn_plugin_malloc(grn_ctx *ctx, size_t size, const char *file, int line,
                   const char *func)
@@ -887,9 +1138,15 @@ grn_plugin_proc_get_var_by_offset(grn_ctx *ctx, grn_user_data *user_data,
 
 const char *
 grn_plugin_win32_base_dir(void)
+{
+  return grn_plugin_windows_base_dir();
+}
+
+const char *
+grn_plugin_windows_base_dir(void)
 {
 #ifdef WIN32
-  return grn_win32_base_dir();
+  return grn_windows_base_dir();
 #else /* WIN32 */
   return NULL;
 #endif /* WIN32 */
